//////////////////////////////////////////////////////////////////////////////
// gloceanconfig.cc
// -------------------
// Config module for GLOcean window decoration
// -------------------
// Copyright (c) 2004 R Paquette <bitwit@digilanti.org>
// Please see the header file for copyright and license information.
//////////////////////////////////////////////////////////////////////////////

#include <kconfig.h>
#include <klocale.h>
#include <kglobal.h>
#include <qbuttongroup.h>
#include <qgroupbox.h>
#include <qradiobutton.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qspinbox.h>
#include <qtoolbutton.h>
#include <qlineedit.h>
#include <qslider.h>
#include <qlabel.h>
#include <qimage.h>
#include <qbitmap.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qfont.h>
#include <qwhatsthis.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <kdecoration.h>

#include "gloceanconfig.h"
#include "configdialog.h"

//////////////////////////////////////////////////////////////////////////////
// GLOceanConfig()
// -------------
// Constructor

GLOceanConfig::GLOceanConfig(KConfig* config, QWidget* parent)
        : QObject(parent), config_(0), dialog_(0) {
    // create the configuration object
    config_ = new KConfig("kwingloceanrc");
    KGlobal::locale()->insertCatalogue("kwin_glocean_config");

    // create and show the configuration dialog
    dialog_ = new ConfigDialog(parent);
    dialog_->show();

    // load the configuration
    load(config_);

    // setup the connections for title align
    connect(dialog_->titlealign, SIGNAL(clicked(int)),this, SLOT(selectionChanged(int)));
    // setup the connections for corner round
    connect(dialog_->topleft, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->topright, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->botleft, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->botright, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    // setup the connections for spin boxes
    connect(dialog_->titlesize, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->buttonsize, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->framesize, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->framedelay, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    // setup connections for filenames
    connect( dialog_->filenamea, SIGNAL( textChanged(const QString &)),this,SLOT(fileChanged(const QString &)));
    connect( dialog_->filenameb, SIGNAL( textChanged(const QString &)),this,SLOT(fileChanged(const QString &)));
    // setup connections for color options
    connect(dialog_->usetitlecolors, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->hidebuttons, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->alphablend, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->buttonblend, SIGNAL(valueChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->textblend, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->activeonly, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->animate, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->toolButton1, SIGNAL(released()),this,SLOT(getFileNameA()));
    connect(dialog_->toolButton2, SIGNAL(released()),this,SLOT(getFileNameB()));
    connect(dialog_->aboutButton, SIGNAL(released()),this,SLOT(aboutButtonReleased()));

    connect(dialog_->alphablend, SIGNAL(valueChanged(int)),this,SLOT(fileaChanged()));
    connect(dialog_->buttonblend, SIGNAL(valueChanged(int)),this,SLOT(filebChanged()));

    connect(dialog_->buttonSlider, SIGNAL(valueChanged(int)),dialog_->buttonblend,SLOT(setValue(int)));
    connect(dialog_->backgroundSlider, SIGNAL(valueChanged(int)),dialog_->alphablend,SLOT(setValue(int)));
    connect(dialog_->symboltype, SIGNAL(activated(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->animtype, SIGNAL(activated(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->fullframe, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->hwrender, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));
    connect(dialog_->shadow, SIGNAL(stateChanged(int)),this,SLOT(selectionChanged(int)));

    preview = new Preview;

    fd = new QFileDialog();
    fd->setContentsPreviewEnabled( TRUE );
    fd->setContentsPreview( preview, preview );
    fd->setPreviewMode( QFileDialog::Contents );
    fd->setFilter ( "Images (*.png *.jpg *.xpm)" );

}

//////////////////////////////////////////////////////////////////////////////
// ~GLOceanConfig()
// --------------
// Destructor

GLOceanConfig::~GLOceanConfig() {
    if (dialog_)
        delete dialog_;
    if (config_)
        delete config_;
}

//////////////////////////////////////////////////////////////////////////////
// selectionChanged()
// ------------------
// Selection has changed
void GLOceanConfig::selectionChanged(int) {
    // don't want buttons bigger than titlebar
    if(dialog_->buttonsize->value() > dialog_->titlesize->value())
        dialog_->buttonsize->setValue(dialog_->titlesize->value());

    symbolChanged();
    
    emit changed();
}

void GLOceanConfig::fileChanged(const QString &) {
    emit changed();
}

void GLOceanConfig::fileaChanged() {

    QImage imt( dialog_->filenamea->text() );
    QImage im(imt.smoothScale(dialog_->aPixmap->width(), dialog_->aPixmap->width()));
    int y = dialog_->aPixmap->width()/2 - dialog_->aPixmap->height()/2;
    QColor c = dialog_->eraseColor();
    QImage imp(fadeImage( im, 1.0-(dialog_->alphablend->value()/100.0), c ));

    QPixmap pm(imp.copy ( 0, y, dialog_->aPixmap->width(), dialog_->aPixmap->height() ) );

   dialog_->aPixmap->setPixmap( pm );
}

void GLOceanConfig::filebChanged() {

    QImage imt( dialog_->filenameb->text() );
    QColor c = dialog_->eraseColor();
    QImage im(fadeImage( imt, 1.0-(dialog_->buttonblend->value()/100.0), c ));
    QPixmap pm(im.smoothScale(24,24));
    dialog_->bPixmap->setPixmap( pm );
}

void GLOceanConfig::symbolChanged()
{
        // setup which bits for symbols
        switch(dialog_->symboltype->currentItem()) {
        case 0:
                setButtonBits(default_bits);
                break;
        case 1:
                setButtonBits(plastic_bits);
                break;
        case 2:
                setButtonBits(kstep_bits);
                break;
        case 3:
                setButtonBits(control_bits);
                break;
        default:
                setButtonBits(default_bits);
                break;
        }
}
//////////////////////////////////////////////////////////////////////////////
// setButtonBits()
// ------------
// Add buttons bits
void GLOceanConfig::setButtonBits( const unsigned char bits[][8] )
{
        QPixmap pm( dialog_->symbols->width(), dialog_->symbols->height() );
        QPainter p( &pm );
	p.fillRect(0,0,dialog_->symbols->width(), dialog_->symbols->height(), dialog_->paletteBackgroundColor() );
        p.setBackgroundColor(dialog_->paletteBackgroundColor() );
	p.setPen(Qt::black);
	int y = 12;

	p.drawPixmap( 0,y, QBitmap(8, 8, bits[0], true) );
	p.drawPixmap( 13,y, QBitmap(8, 8, bits[2], true) );
	p.drawPixmap( 26,y, QBitmap(8, 8, bits[3], true) );
	p.drawPixmap( 39,y, QBitmap(8, 8, bits[4], true) );
	p.drawPixmap( 52,y, QBitmap(8, 8, bits[5], true) );
	p.drawPixmap( 65,y, QBitmap(8, 8, bits[6], true) );
	y += 14;
	p.drawPixmap( 0,y, QBitmap(8, 8, bits[7], true) );
	p.drawPixmap( 13,y, QBitmap(8, 8, bits[8], true) );
	p.drawPixmap( 26,y, QBitmap(8, 8, bits[9], true) );
	p.drawPixmap( 39,y, QBitmap(8, 8, bits[10], true) );
	p.drawPixmap( 52,y, QBitmap(8, 8, bits[11], true) );
	p.drawPixmap( 65,y, QBitmap(8, 8, bits[12], true) );
        
	dialog_->symbols->setPixmap(pm);
}
//////////////////////////////////////////////////////////////////////////////
// load()
// ------
// Load configuration data

void GLOceanConfig::load(KConfig*) {
    config_->setGroup("General");

    QString value = config_->readEntry("TitleAlignment", "AlignHCenter");
    QRadioButton *button = (QRadioButton*)dialog_->titlealign->child(value);
    if (button)
        button->setChecked(true);

    int cornersFlag = config_->readNumEntry("RoundCorners",1 );
    dialog_->topleft->setChecked( cornersFlag & 1 );
    dialog_->topright->setChecked( cornersFlag & 2 );
    dialog_->botleft->setChecked( cornersFlag & 4 );
    dialog_->botright->setChecked( cornersFlag & 8 );

    dialog_->titlesize->setValue( config_->readNumEntry("TitleSize",24 ) );
    dialog_->buttonsize->setValue( config_->readNumEntry("ButtonSize",16 ) );
    dialog_->framesize->setValue( config_->readNumEntry("FrameSize",1 ) );
    dialog_->framedelay->setValue( config_->readNumEntry("FrameDelay",60 ) );
    dialog_->filenamea->setText( config_->readEntry( "Background", ".glocean/background.jpg" ) );
    dialog_->filenameb->setText( config_->readEntry( "Button", ".glocean/button.png" ) );
    dialog_->usetitlecolors->setChecked( config_->readBoolEntry("UseTitleColors",false) );
    dialog_->hidebuttons->setChecked( config_->readBoolEntry("HideButtons",false) );
    dialog_->alphablend->setValue( config_->readNumEntry("AlphaBlend",50 ) );
    dialog_->buttonblend->setValue( config_->readNumEntry("ButtonBlend",50 ) );

    dialog_->textblend->setChecked( config_->readBoolEntry("FadeText",true) );
    dialog_->activeonly->setChecked( config_->readBoolEntry("ActiveOnly",false) );
    dialog_->animate->setChecked( config_->readBoolEntry("Animate",false) );

    QImage imt( dialog_->filenamea->text() );
    QImage im(imt.smoothScale(dialog_->aPixmap->width(), dialog_->aPixmap->width()));

    dialog_->aPixmap->pixmap()->convertFromImage(im.copy ( 0, dialog_->aPixmap->width()/2 - dialog_->aPixmap->height()/2,
    dialog_->aPixmap->width(),
    dialog_->aPixmap->height() ) );
    dialog_->backgroundSlider->setValue(config_->readNumEntry("AlphaBlend",50 ));
    dialog_->buttonSlider->setValue(config_->readNumEntry("ButtonBlend",50 ));

    dialog_->symboltype->setCurrentItem(config_->readNumEntry("SymbolType",0 ));
    dialog_->animtype->setCurrentItem(config_->readNumEntry("AnimType",3 ));
    dialog_->fullframe->setChecked( config_->readBoolEntry("FullFrame",false) );
    dialog_->hwrender->setChecked( config_->readBoolEntry("HWRender",false) );
    dialog_->shadow->setChecked( config_->readBoolEntry("CaptionShadow",false) );
    
    fileaChanged();
    filebChanged();
    symbolChanged();
}

//////////////////////////////////////////////////////////////////////////////
// save()
// ------
// Save configuration data

void GLOceanConfig::save(KConfig*) {
    config_->setGroup("General");

    QRadioButton *button = (QRadioButton*)dialog_->titlealign->selected();
    if (button)
        config_->writeEntry("TitleAlignment", QString(button->name()));

    int cornersFlag = 0;
    if(dialog_->topleft->isChecked())
        cornersFlag |= 1;
    if(dialog_->topright->isChecked())
        cornersFlag |= 2;
    if(dialog_->botleft->isChecked())
        cornersFlag |= 4;
    if(dialog_->botright->isChecked())
        cornersFlag |= 8;
    config_->writeEntry("RoundCorners", cornersFlag );

    config_->writeEntry("TitleSize",  dialog_->titlesize->value() );
    config_->writeEntry("ButtonSize", dialog_->buttonsize->value() );
    config_->writeEntry("FrameSize",  dialog_->framesize->value() );
    config_->writeEntry("FrameDelay", dialog_->framedelay->value() );
    config_->writeEntry("Background", dialog_->filenamea->text() );
    config_->writeEntry("Button",     dialog_->filenameb->text() );

    config_->writeEntry("UseTitleColors", dialog_->usetitlecolors->isChecked());
    config_->writeEntry("HideButtons",   dialog_->hidebuttons->isChecked());
    config_->writeEntry("AlphaBlend",   dialog_->alphablend->value());
    config_->writeEntry("ButtonBlend",   dialog_->buttonblend->value());
    config_->writeEntry("FadeText",   dialog_->textblend->isChecked());
    config_->writeEntry("ActiveOnly",   dialog_->activeonly->isChecked());
    config_->writeEntry("Animate",   dialog_->animate->isChecked());
    config_->writeEntry("SymbolType", dialog_->symboltype->currentItem());
    config_->writeEntry("AnimType", dialog_->animtype->currentItem());
    config_->writeEntry("FullFrame",   dialog_->fullframe->isChecked());
    config_->writeEntry("HWRender",   dialog_->hwrender->isChecked());
    config_->writeEntry("CaptionShadow",   dialog_->shadow->isChecked());
    config_->sync();
}

//////////////////////////////////////////////////////////////////////////////
// defaults()
// ----------
// Set configuration defaults

void GLOceanConfig::defaults() {
    QRadioButton *button =
        (QRadioButton*)dialog_->titlealign->child("AlignHCenter");
    if (button)
        button->setChecked(true);

    int cornersFlag = config_->readNumEntry("RoundCorners",1 );
    dialog_->topleft->setChecked( cornersFlag & 1 );
    dialog_->topright->setChecked( cornersFlag & 2 );
    dialog_->botleft->setChecked( cornersFlag & 4 );
    dialog_->botright->setChecked( cornersFlag & 8 );

    dialog_->titlesize->setValue(24);
    dialog_->buttonsize->setValue(16);
    dialog_->framesize->setValue(1);
    dialog_->framedelay->setValue(90);
    dialog_->filenamea->setText( ".glocean/" );
    dialog_->filenameb->setText( ".glocean/" );

    dialog_->usetitlecolors->setChecked(false);
    dialog_->hidebuttons->setChecked(false);
    dialog_->alphablend->setValue( 50 );
    dialog_->buttonblend->setValue( 50 );
    dialog_->backgroundSlider->setValue(config_->readNumEntry("AlphaBlend",50 ));
    dialog_->buttonSlider->setValue(config_->readNumEntry("ButtonBlend",50 ));
    dialog_->textblend->setChecked(true);
    dialog_->activeonly->setChecked(false);
    dialog_->animate->setChecked(false);
    dialog_->symboltype->setCurrentItem(0);
    dialog_->animtype->setCurrentItem(3);
    dialog_->fullframe->setChecked(false);
    dialog_->hwrender->setChecked(false);
    dialog_->shadow->setChecked(false);
    
    symbolChanged();
}

void GLOceanConfig::getFileNameA()
{
    fd->setSelection(dialog_->filenamea->text());
    preview->previewUrl(dialog_->filenamea->text());
    if ( fd->exec() == QDialog::Accepted ) {
        dialog_->filenamea->setText( fd->selectedFile() );
        fileaChanged();
        emit changed();
    }

}

void GLOceanConfig::getFileNameB()
{
    fd->setSelection(dialog_->filenameb->text());
    preview->previewUrl(dialog_->filenameb->text());
    if ( fd->exec() == QDialog::Accepted ) {
        dialog_->filenameb->setText( fd->selectedFile() );
        filebChanged();
        emit changed();
    }
}


void GLOceanConfig::aboutButtonReleased()
{
    QString mess("Author: bitwit@digilanti.org\n\
Copyright 2004 Rick Paquette\n\n\
GPL  Free software for free minds.\n\n\
V 1.1 Release 26\n\n\
Thank you Sascha, for all your help\n\
and the great GLFont class.");

    QMessageBox::about( dialog_, "GLOcean", mess );
}

//from kimageefect.cpp
QImage& GLOceanConfig::fadeImage(QImage &img, float val, const QColor &color)
{
    if (img.width() == 0 || img.height() == 0)
      return img;

    // We don't handle bitmaps
    if (img.depth() == 1)
    return img;

    unsigned char tbl[256];
    for (int i=0; i<256; i++)
    tbl[i] = (int) (val * i + 0.5);

    int red = color.red();
    int green = color.green();
    int blue = color.blue();

    QRgb col;
    int r, g, b, cr, cg, cb;

    if (img.depth() <= 8) {
    // pseudo color
    for (int i=0; i<img.numColors(); i++) {
        col = img.color(i);
        cr = qRed(col); cg = qGreen(col); cb = qBlue(col);
        if (cr > red)
        r = cr - tbl[cr - red];
        else
        r = cr + tbl[red - cr];
        if (cg > green)
        g = cg - tbl[cg - green];
        else
        g = cg + tbl[green - cg];
        if (cb > blue)
        b = cb - tbl[cb - blue];
        else
        b = cb + tbl[blue - cb];
        img.setColor(i, qRgba(r, g, b, qAlpha(col)));
    }

    } else {
    // truecolor
        for (int y=0; y<img.height(); y++) {
            QRgb *data = (QRgb *) img.scanLine(y);
            for (int x=0; x<img.width(); x++) {
                col = *data;
                cr = qRed(col); cg = qGreen(col); cb = qBlue(col);
                if (cr > red)
                    r = cr - tbl[cr - red];
                else
                    r = cr + tbl[red - cr];
                if (cg > green)
                    g = cg - tbl[cg - green];
                else
                    g = cg + tbl[green - cg];
                if (cb > blue)
                    b = cb - tbl[cb - blue];
                else
                    b = cb + tbl[blue - cb];
                *data++ = qRgba(r, g, b, qAlpha(col));
            }
        }
    }
     return img;
}

//////////////////////////////////////////////////////////////////////////////
// Plugin Stuff                                                             //
//////////////////////////////////////////////////////////////////////////////

extern "C" {
    QObject* allocate_config(KConfig* config, QWidget* parent) {
        return (new GLOceanConfig(config, parent));
    }
}

#include "gloceanconfig.moc"

