//////////////////////////////////////////////////////////////////////////////
// gloceanclient.cc
// -------------------
// GLOcean window decoration for KDE
// -------------------
// Copyright (c) 2004 R Paquette
// Please see the header file for copyright and license information.
//////////////////////////////////////////////////////////////////////////////

#include <kconfig.h>
#include <kdeversion.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <klocale.h>

#include <qtimer.h>
#include <qbitmap.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qtooltip.h>
#include "gloceanclient.h"

#include <GL/glx.h>

using namespace GLOcean;

GLOceanFactory* factory=NULL;

extern "C" KDecorationFactory* create_factory()
{
        return new GLOcean::GLOceanFactory();
}

//////////////////////////////////////////////////////////////////////////////
// GLOceanFactory Class                                                     //
//////////////////////////////////////////////////////////////////////////////
GLXContext GLOceanFactory::glxcontext;
bool GLOceanFactory::initialized_              = false;
Qt::AlignmentFlags GLOceanFactory::titlealign_ = Qt::AlignHCenter;
int GLOceanFactory::cornerflags_               = 1;
int GLOceanFactory::titlesize_                 = 24;
int GLOceanFactory::buttonsize_                = 15;
int GLOceanFactory::framesize_                 = 1;
int GLOceanFactory::framedelay_                = 90;
bool GLOceanFactory::titlecolors_              = false;
bool GLOceanFactory::hidebuttons_              = false;
int GLOceanFactory::alphablend_                = 70;
int GLOceanFactory::buttonblend_               = 70;
bool GLOceanFactory::fadetext_                 = false;
bool GLOceanFactory::activeonly_               = true;
bool GLOceanFactory::animate_                  = false;
bool GLOceanFactory::glinitialized_            = false;
int GLOceanFactory::symboltype_                = 0;
int GLOceanFactory::animtype_                  = 3;
bool GLOceanFactory::fullframe_                = false;
bool GLOceanFactory::hwrender_                 = false;
bool GLOceanFactory::shadow_                   = false;
QImage GLOceanFactory::backfile;
QImage GLOceanFactory::buttfile;

//////////////////////////////////////////////////////////////////////////////
// GLOceanFactory()
// ----------------
// Constructor

GLOceanFactory::GLOceanFactory()
{
        initialized_ = readConfig();
        ::factory=this;
        glinitialized_=false;
        init();
}

//////////////////////////////////////////////////////////////////////////////
// ~GLOceanFactory()
// -----------------
// Destructor

GLOceanFactory::~GLOceanFactory()
{
        initialized_ = false;
        glinitialized_=false;
        if (glxcontext)
                glXDestroyContext(qt_xdisplay(),glxcontext);
}

/////////////////////////////////////////////////////////////
// Initialize GLX
// -----------------
//
void GLOceanFactory::init()
{

        Display *dpy=qt_xdisplay();
        int attrib[] = { GLX_RGBA, None }; // , GLX_DOUBLEBUFFER, GLX_DEPTH_SIZE, 1
        int scrnum = DefaultScreen( dpy );

        XVisualInfo *visinfo = glXChooseVisual( dpy, scrnum, attrib );
        if (visinfo) {
                glxcontext = glXCreateContext( dpy, visinfo, NULL, true );
                if (!glxcontext) {
                        qDebug("Error: glXCreateContext failed\n");
                } else
                        glinitialized_ = true;
                XFree( visinfo );
        } else
                qDebug("Error: couldn't get a RGBA visual\n");
}

//////////////////////////////////////////////////////////////////////////////
// createDecoration()
// -----------------
// Create the decoration

KDecoration* GLOceanFactory::createDecoration(KDecorationBridge* b)
{
        return new GLOceanClient(b, this);
}

//////////////////////////////////////////////////////////////////////////////
// reset()
// -------
// Reset the handler. Returns true if decorations need to be remade, false if
// only a repaint is necessary

bool GLOceanFactory::reset(unsigned long changed)
{
        initialized_ = false;
        bool confchange = readConfig();
        initialized_ = true;

        if (confchange ||
                        (changed & (SettingDecoration | SettingButtons | SettingBorder))) {
                return true;
        } else {
                resetDecorations(changed);
                return false;
        }
}

//////////////////////////////////////////////////////////////////////////////
// readConfig()
// ------------
// Read in the configuration file

bool GLOceanFactory::readConfig()
{
        // create a config object
        KConfig config("kwingloceanrc");
        config.setGroup("General");

        QString value = config.readEntry("TitleAlignment", "AlignHCenter");
        if (value == "AlignLeft")
                titlealign_ = Qt::AlignLeft;
        else if (value == "AlignHCenter")
                titlealign_ = Qt::AlignHCenter;
        else if (value == "AlignRight")
                titlealign_ = Qt::AlignRight;

        cornerflags_=config.readNumEntry("RoundCorners",1);
        titlesize_=config.readNumEntry("TitleSize",24);
        buttonsize_=config.readNumEntry("ButtonSize",16);
        framesize_=config.readNumEntry("FrameSize",1);
        framedelay_=config.readNumEntry("FrameDelay",60);

        backfile.load(config.readEntry("Background",".glocean/background.jpg"));
        buttfile.load(config.readEntry("Button",".glocean/button.png"));

        GlobeWidget::imageBack = backfile;
        GlobeWidget::imageButt = buttfile;

        if(buttonsize_ > titlesize_ )
                buttonsize_ = titlesize_;

        titlecolors_=config.readBoolEntry("UseTitleColors",false);
        hidebuttons_=config.readBoolEntry("HideButtons",false);
        alphablend_=config.readNumEntry("AlphaBlend",70);
        buttonblend_=config.readNumEntry("ButtonBlend",70);
        fadetext_=config.readBoolEntry("FadeText",true);
        activeonly_=config.readBoolEntry("ActiveOnly",false);
        animate_=config.readBoolEntry("Animate",false);
        symboltype_=config.readNumEntry("SymbolType",0);
        animtype_=config.readNumEntry("AnimType",3);
        fullframe_=config.readBoolEntry("FullFrame",false);
        hwrender_=config.readBoolEntry("HWRender",false);
        shadow_=config.readBoolEntry("CaptionShadow",false);
        GlobeWidget::usetitlecolor = titlecolors_;
        GlobeWidget::alphablend = alphablend_;
        GlobeWidget::buttonblend = buttonblend_;

        return true;
}

//////////////////////////////////////////////////////////////////////////////
// GLOceanButton Class                                                      //
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// GLOceanButton()
// ---------------
// Constructor

GLOceanButton::GLOceanButton(GLOceanClient *parent, const char *name,
                             const QString& tip, ButtonType type,
                             const unsigned char *bitmap)
                : QButton(parent->widget(), name), client_(parent), type_(type),
                deco_(0), lastmouse_(NoButton), hover_(false)
{
        setBackgroundMode(NoBackground);
        setFixedSize( ::factory->buttonSize(), ::factory->buttonSize());
        setCursor(arrowCursor);
        if (bitmap)
                setBitmap(bitmap);
        QToolTip::add
                (this, tip);
}

GLOceanButton::~GLOceanButton()
{
        if (deco_)
                delete deco_;
}

//////////////////////////////////////////////////////////////////////////////
// setBitmap()
// -----------
// Set the button decoration

void GLOceanButton::setBitmap(const unsigned char *bitmap)
{
        if (!bitmap)
                return;
        if (deco_)
                delete deco_;
        deco_ = new QBitmap(8, 8, bitmap, true);
        deco_->setMask(*deco_);
        repaint(false);
}

//////////////////////////////////////////////////////////////////////////////
// sizeHint()
// ----------
// Return size hint

QSize GLOceanButton::sizeHint() const
{
        return QSize(::factory->buttonSize(), ::factory->buttonSize());
}

//////////////////////////////////////////////////////////////////////////////
// mousePressEvent()
// -----------------
// Button has been pressed

void GLOceanButton::mousePressEvent(QMouseEvent* e)
{
        lastmouse_ = e->button();
        int button = LeftButton;
        if ((type_ != ButtonMax) && (e->button() != LeftButton)) {
                button = NoButton;
        }
        QMouseEvent me(e->type(), e->pos(), e->globalPos(),
                       button, e->state());
        QButton::mousePressEvent(&me);
}

//////////////////////////////////////////////////////////////////////////////
// mouseReleaseEvent()
// -----------------
// Button has been released

void GLOceanButton::mouseReleaseEvent(QMouseEvent* e)
{
        lastmouse_ = e->button();
        int button = LeftButton;
        if ((type_ != ButtonMax) && (e->button() != LeftButton))
                button = NoButton;
        QMouseEvent me(e->type(), e->pos(), e->globalPos(), button, e->state());
        QButton::mouseReleaseEvent(&me);
}

//////////////////////////////////////////////////////////////////////////////
// drawButton()
// ------------
// Draw the button

void GLOceanButton::drawButton(QPainter *painter)
{

        if (! ::factory->initialized())
                return;

        int dx, dy;

        QColorGroup group = KDecoration::options()->colorGroup(KDecoration::ColorTitleBar, client_->isActive());
        if (type_ == ButtonMenu) {
                dx = (width() - 16) / 2;
                dy = (height() - 16) / 2;
                if (isDown()) {
                        dx++;
                        dy++;
                }
                QPixmap pmcopy(client_->icon().pixmap( QIconSet::Small, hover_ ? QIconSet::Active : QIconSet::Normal));
                if(pmcopy.mask()) {
                        QRegion r1( QRect(0,0,16,16) );
                        r1 &= *pmcopy.mask();
                        r1.translate(dx,dy);
                        setMask( r1 );
                }
                painter->drawPixmap( QRect(dx,dy,16,16), pmcopy);
        } else
                if (deco_) {
                        dx = (width() - 8) / 2;
                        dy = (height() - 8) / 2;
                        if (isDown()) {
                                dx++;
                                dy++;
                        }

                        QRegion r3( 0,0,8,8 );
                        r3 &= *deco_;
                        r3.translate(dx,dy);
                        setMask(r3);

                        hover_ ? painter->setPen(group.light()) : painter->setPen(group.mid()) ;
                        painter->drawPixmap(dx, dy, *deco_);
                }
}

//////////////////////////////////////////////////////////////////////////////
// GLOceanClient Class                                                      //
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// GLOceanClient()
// ---------------
// Constructor

GLOceanClient::GLOceanClient(KDecorationBridge *b, KDecorationFactory *f)
                : KDecoration(b, f)
{
        glGlobe = NULL;
        started_ = false;
}

GLOceanClient::~GLOceanClient()
{}

//////////////////////////////////////////////////////////////////////////////
// init()
// ------
// Actual initializer for class

void GLOceanClient::init()
{
        createMainWidget(WResizeNoErase | WRepaintNoErase);
        widget()->installEventFilter(this);
        widget()->setBackgroundMode(NoBackground);
        ///    widget()->setWindowOpacity ( ::factory->alphaBlend() );
        ///    QImage backImage = QPixmap::grabWindow(QApplication::desktop()->winId());
        ///    widget()->setPaletteBackgroundPixmap( backImage.copy( QRect(widget()->pos(), widget()->size()) ));

        // to offset buttons if smaller so they are centered nicely
        int spc=((::factory->titleSize() - ::factory->buttonSize() - 1)/2);
        // setup layout
        QGridLayout *mainlayout = new QGridLayout(widget(), 4, 3 ); // 4x3 grid
        QHBoxLayout *titlelayout = new QHBoxLayout();
        titlebar_ = new QSpacerItem(1, ::factory->titleSize(), QSizePolicy::Expanding, QSizePolicy::Fixed);
        mainlayout->setResizeMode(QLayout::FreeResize);
        //       mainlayout->addRowSpacing(0, 1);   // top
        mainlayout->addRowSpacing(3, 1);    // bottom
        mainlayout->addColSpacing(0, 1);    // left
        mainlayout->addColSpacing(2, 1);    // right

        // gl stuff
        if( ::factory->glInitialized() ) {
                QGLFormat f;
                f.setDirectRendering( ::factory->hwRender() );
                f.setRgba( TRUE );
                QGLFormat::setDefaultFormat( f );
                // set these static vars before calling new
                GlobeWidget::imageBack = ::factory->backFile();
                GlobeWidget::imageButt = ::factory->buttFile();
                GlobeWidget::usetitlecolor = ::factory->titleColors();
                GlobeWidget::alphablend = ::factory->alphaBlend();
                GlobeWidget::buttonblend = ::factory->buttonBlend();
                // new GL scene for this widget
                glGlobe = new GlobeWidget( widget() );
                glGlobe->installEventFilter(this);
                // setup the GL buttons before call to show
                glGlobe->setHideButtons(::factory->hideButtons());
                glGlobe->setHasHelp(providesContextHelp());
                glGlobe->setHasMin(isMinimizable());
                glGlobe->setHasMax(isMaximizable());
                glGlobe->setHasClose(isCloseable());
                glGlobe->setHasShade(isShadeable());
                glGlobe->setButtonSize(::factory->buttonSize());
                glGlobe->setTitleAlign(::factory->titleAlign());
                glGlobe->setHasTextFade(::factory->fadeText() );
                glGlobe->setHasShadow(::factory->hasShadow() );
                glGlobe->setTitleText(caption(),::factory->titleSize(), ::factory->frameSize());
                glGlobe->setBlendColor( options()->color(ColorTitleBlend, isActive()) );
                glGlobe->setAnimate( ::factory->animateButtons() );
                glGlobe->setAnimType( ::factory->animType() );
                glGlobe->setBackgroundColor( options()->color(ColorTitleBar, isActive() ) );
                glGlobe->setCaptionColor( options()->color(ColorFont, isActive()) );
        }

        mainlayout->addLayout(titlelayout, 1, 1);
        // the window should stretch
        mainlayout->setRowStretch(2, 10);
        mainlayout->setColStretch(1, 10);
        // message in preview widget
        if (isPreview()) {
                QString s;
                if(::factory->glInitialized()) {
                        int major,minor;
                        glXQueryVersion( qt_xdisplay(), &major, &minor );
                        s.sprintf("<center>GLOcean Preview<br>GLX Version:<b>%i.%i</b><br>Direct Render:<b>%s</b></center>",
                                  major,minor,
                                  glXIsDirect( qt_xdisplay(), ::factory->glxContext() ) && ::factory->hwRender() ? "Yes" : "No");
                } else {
                        s.sprintf("<center>GLOcean Preview<br><br><b>No GL !</b></center>");
                }
                mainlayout->addWidget( new QLabel(i18n( s ),widget()), 2, 1, Qt::AlignHCenter | Qt::AlignVCenter);
        } else {
                mainlayout->addItem(new QSpacerItem(0, 0), 2, 1);
        }

        // reset titlebar buttons
        for (int n=0; n<ButtonTypeCount; n++)
                button[n] = 0;
        // setup which bits for symbols
        switch(::factory->symbolType()) {
        case 0:
                setButtonBits(default_bits);
                break;
        case 1:
                setButtonBits(plastic_bits);
                break;
        case 2:
                setButtonBits(kstep_bits);
                break;
        case 3:
                setButtonBits(control_bits);
                break;
        default:
                setButtonBits(default_bits);
                break;
        }

        if(spc>0)
                titlelayout->addSpacing( spc );
        addButtons(titlelayout, options()->titleButtonsLeft());
        titlelayout->addItem(titlebar_);
        addButtons(titlelayout, options()->titleButtonsRight());
        if(spc>0)
                titlelayout->addSpacing( spc );

        // connections
        connect(this, SIGNAL(keepAboveChanged(bool)), SLOT(keepAboveChange(bool)));
        connect(this, SIGNAL(keepBelowChanged(bool)), SLOT(keepBelowChange(bool)));
}

//////////////////////////////////////////////////////////////////////////////
// setButtonBits()
// ------------
// Add buttons bits
void GLOceanClient::setButtonBits( const unsigned char bits[][8] )
{
        close_bits =  bits[0];
        help_bits  = bits[1];
        max_bits  = bits[2];
        min_bits  = bits[3];
        restore_bits  = bits[4];
        sticky_on_bits  = bits[5];
        sticky_off_bits  = bits[6];
        above_on_bits  = bits[7];
        above_off_bits  = bits[8];
        below_on_bits  = bits[9];
        below_off_bits  = bits[10];
        shade_on_bits  = bits[11];
        shade_off_bits  =bits[12];
}
//////////////////////////////////////////////////////////////////////////////
// addButtons()
// ------------
// Add buttons to title layout

void GLOceanClient::addButtons(QBoxLayout *layout, const QString& s)
{
        const unsigned char *bitmap;
        QString tip;
        if (s.length() > 0) {
                for (unsigned n=0; n < s.length(); n++) {
                        switch (s[n]) {
                        case 'M': // Menu button
                                if (!button[ButtonMenu]) {
                                        button[ButtonMenu] =
                                                new GLOceanButton(this, "menu", i18n("Menu"),
                                                                  ButtonMenu, 0);
                                        connect(button[ButtonMenu], SIGNAL(pressed()),
                                                this, SLOT(menuButtonPressed()));
                                        layout->addWidget(button[ButtonMenu]);
                                }
                                break;

                        case 'S': // Sticky button
                                if (!button[ButtonSticky]) {
                                        if (isOnAllDesktops()) {
                                                bitmap = sticky_on_bits;
                                                tip = i18n("Un-Sticky");
                                        } else {
                                                bitmap = sticky_off_bits;
                                                tip = i18n("Sticky");
                                        }
                                        button[ButtonSticky] =
                                                new GLOceanButton(this, "sticky", tip,
                                                                  ButtonSticky, bitmap);
                                        connect(button[ButtonSticky], SIGNAL(clicked()),
                                                this, SLOT(toggleOnAllDesktops()));
                                        layout->addWidget(button[ButtonSticky]);
                                }
                                break;

                        case 'H': // Help button
                                if ((!button[ButtonHelp]) && providesContextHelp()) {
                                        button[ButtonHelp] =
                                                new GLOceanButton(this, "help", i18n("Help"),
                                                                  ButtonHelp, help_bits);
                                        connect(button[ButtonHelp], SIGNAL(clicked()),
                                                this, SLOT(showContextHelp()));
                                        layout->addWidget(button[ButtonHelp]);
                                }
                                break;

                        case 'I': // Minimize button
                                if ((!button[ButtonMin]) && isMinimizable())  {
                                        button[ButtonMin] =
                                                new GLOceanButton(this, "iconify", i18n("Minimize"),
                                                                  ButtonMin, min_bits);
                                        connect(button[ButtonMin], SIGNAL(clicked()),
                                                this, SLOT(minimize()));
                                        layout->addWidget(button[ButtonMin]);
                                }
                                break;

                        case 'A': // Maximize button
                                if ((!button[ButtonMax]) && isMaximizable()) {
                                        if (maximizeMode() == MaximizeFull) {
                                                bitmap = restore_bits;
                                                tip = i18n("Restore");
                                        } else {
                                                bitmap = max_bits;
                                                tip = i18n("Maximize");
                                        }
                                        button[ButtonMax]  =
                                                new GLOceanButton(this, "maximize", tip,
                                                                  ButtonMax, bitmap);
                                        connect(button[ButtonMax], SIGNAL(clicked()),
                                                this, SLOT(maxButtonPressed()));
                                        layout->addWidget(button[ButtonMax]);
                                }
                                break;

                        case 'X': // Close button
                                if ((!button[ButtonClose]) && isCloseable()) {
                                        button[ButtonClose] =
                                                new GLOceanButton(this, "close", i18n("Close"),
                                                                  ButtonClose, close_bits);
                                        connect(button[ButtonClose], SIGNAL(clicked()),
                                                this, SLOT(closeWindow()));
                                        layout->addWidget(button[ButtonClose]);
                                }
                                break;

                        case 'F': // Above button
                                if ((!button[ButtonAbove])) {
                                        button[ButtonAbove] =
                                                new GLOceanButton(this, "above",
                                                                  i18n("Keep Above Others"), ButtonAbove,
                                                                  keepAbove() ? above_on_bits : above_off_bits);
                                        connect(button[ButtonAbove], SIGNAL(clicked()),
                                                this, SLOT(aboveButtonPressed()));
                                        layout->addWidget(button[ButtonAbove]);
                                }
                                break;

                        case 'B': // Below button
                                if ((!button[ButtonBelow])) {
                                        button[ButtonBelow] =
                                                new GLOceanButton(this, "below",
                                                                  i18n("Keep Below Others"), ButtonBelow,
                                                                  keepBelow() ? below_on_bits : below_off_bits);
                                        connect(button[ButtonBelow], SIGNAL(clicked()),
                                                this, SLOT(belowButtonPressed()));
                                        layout->addWidget(button[ButtonBelow]);
                                }
                                break;

                        case 'L': // Shade button
                                if ((!button[ButtonShade && isShadeable()])) {
                                        if ( isSetShade()) {
                                                bitmap = shade_on_bits;
                                                tip = i18n("Unshade");
                                        } else {
                                                bitmap = shade_off_bits;
                                                tip = i18n("Shade");
                                        }
                                        button[ButtonShade] =
                                                new GLOceanButton(this, "shade", tip,
                                                                  ButtonShade, bitmap);
                                        connect(button[ButtonShade], SIGNAL(clicked()),
                                                this, SLOT(shadeButtonPressed()));
                                        layout->addWidget(button[ButtonShade]);
                                }
                                break;

                        case '_': // Spacer item
                                layout->addSpacing(::factory->buttonSize()/2);
                        }
                }
        }
}

//////////////////////////////////////////////////////////////////////////////
// enterEvent()
// --------------
// window entered
void GLOceanClient::enterEvent()
{
        if(::factory->glInitialized()) {
                glGlobe->view_rotz=0.0;
                if(::factory->hideButtons())
                        glGlobe->lower();
        }
}

//////////////////////////////////////////////////////////////////////////////
// enterEvent()
// --------------
// window left
void GLOceanClient::leaveEvent()
{
        if(::factory->hideButtons()) {
                if(::factory->glInitialized())
                        glGlobe->raise();
                button[ButtonMenu]->raise();
        }
}

//////////////////////////////////////////////////////////////////////////////
// activeChange()
// --------------
// window active state has changed

void GLOceanClient::activeChange()
{
        for (int n=0; n<ButtonTypeCount; n++)
                if (button[n] )
                        button[n]->reset();
        if(::factory->glInitialized()) {
                glGlobe->setHideButtons( ::factory->hideButtons() );
                glGlobe->setBackgroundColor( options()->color(ColorTitleBar, isActive()) );
                glGlobe->setCaptionColor( options()->color(ColorFont, isActive()) );
                glGlobe->setBlendColor( options()->color(ColorTitleBlend, isActive()) );
                if( ::factory->activeOnly() ) {
                        if( isActive() && !started_ ) {
                                glGlobe->start(::factory->frameDel());
                                started_ = true;
                        }
                        if( !isActive() && started_) {
                                glGlobe->stop();
                                started_ = false;
                        }
                }
                if(isPreview())
                        glGlobe->setTitleChange( true );
        }
        widget()->repaint(false);
}

//////////////////////////////////////////////////////////////////////////////
// captionChange()
// ---------------
// The title has changed

void GLOceanClient::captionChange()
{
        if(::factory->glInitialized()) {
                glGlobe->setTitleText(caption(), ::factory->titleSize(), ::factory->frameSize());
                if(!isPreview())
                        glGlobe->setTitleChange( true );
        }
        widget()->repaint(titlebar_->geometry(), false);
}

//////////////////////////////////////////////////////////////////////////////
// desktopChange()
// ---------------
// Called when desktop/sticky changes

void GLOceanClient::desktopChange()
{
        bool d = isOnAllDesktops();
        if (button[ButtonSticky]) {
                button[ButtonSticky]->setBitmap(d ? sticky_on_bits : sticky_off_bits);
                QToolTip::remove
                        (button[ButtonSticky]);
                QToolTip::add
                        (button[ButtonSticky], d ? i18n("Un-Sticky") : i18n("Sticky"));
        }
}

//////////////////////////////////////////////////////////////////////////////
// iconChange()
// ------------
// The Icon has changed

void GLOceanClient::iconChange()
{
        if (button[ButtonMenu]) {
                button[ButtonMenu]->setBitmap(0);
                button[ButtonMenu]->repaint(false);
        }
}

//////////////////////////////////////////////////////////////////////////////
// maximizeChange()
// ----------------
// Maximized state has changed

void GLOceanClient::maximizeChange()
{
        bool m = (maximizeMode() == MaximizeFull);
        if (button[ButtonMax]) {
                button[ButtonMax]->setBitmap(m ? restore_bits : max_bits);
                QToolTip::remove
                        (button[ButtonMax]);
                QToolTip::add
                        (button[ButtonMax], m ? i18n("Restore") : i18n("Maximize"));
        }
}

//////////////////////////////////////////////////////////////////////////////
// shadeChange()
// -------------
// Called when window shading changes

void GLOceanClient::shadeChange()
{
        bool s = isSetShade();
        if (button[ButtonShade]) {
                button[ButtonShade]->setBitmap(s ? shade_on_bits : shade_off_bits);
                QToolTip::remove
                        (button[ButtonShade]);
                QToolTip::add
                        (button[ButtonShade], s ? i18n("Unshade") : i18n("Shade"));
        }
}

//////////////////////////////////////////////////////////////////////////////
// keepAboveChange()
// ------------
// The above state has changed

void GLOceanClient::keepAboveChange(bool a)
{
        if (button[ButtonAbove]) {
                button[ButtonAbove]->setBitmap(a ? above_on_bits : above_off_bits);
        }
}

//////////////////////////////////////////////////////////////////////////////
// keepBelowChange()
// ------------
// The below state has changed

void GLOceanClient::keepBelowChange(bool b)
{
        if (button[ButtonBelow]) {
                button[ButtonBelow]->setBitmap(b ? below_on_bits : below_off_bits);
        }
}

//////////////////////////////////////////////////////////////////////////////
// borders()
// ----------
// Get the size of the borders

void GLOceanClient::borders(int &l, int &r, int &t, int &b) const
{
        l = r = ::factory->frameSize();
        t = ::factory->titleSize() + ::factory->frameSize();
        b = ::factory->frameSize() * 2;
}

//////////////////////////////////////////////////////////////////////////////
// resize()
// --------
// Called to resize the window

void GLOceanClient::resize(const QSize &size)
{
        widget()->resize(size);
}

//////////////////////////////////////////////////////////////////////////////
// minimumSize()
// -------------
// Return the minimum allowable size for this window

QSize GLOceanClient::minimumSize() const
{
        return widget()->minimumSize();
}

//////////////////////////////////////////////////////////////////////////////
// mousePosition()
// ---------------
// Return logical mouse position

KDecoration::Position GLOceanClient::mousePosition(const QPoint &point) const
{
        const int corner = 20;
        Position pos;
        int fs = ::factory->frameSize();

        if (point.y() <= fs) {
                // inside top frame
                if (point.x() <= corner)
                        pos = PositionTopLeft;
                else if (point.x() >= (width()-corner))
                        pos = PositionTopRight;
                else
                        pos = PositionTop;
        } else if (point.y() >= (height()-fs*2)) {
                // inside handle
                if (point.x() <= corner)
                        pos = PositionBottomLeft;
                else if (point.x() >= (width()-corner))
                        pos = PositionBottomRight;
                else
                        pos = PositionBottom;
        } else if (point.x() <= fs ) {
                // on left frame
                if (point.y() <= corner)
                        pos = PositionTopLeft;
                else if (point.y() >= (height()-corner))
                        pos = PositionBottomLeft;
                else
                        pos = PositionLeft;
        } else if (point.x() >= width()-fs) {
                // on right frame
                if (point.y() <= corner)
                        pos = PositionTopRight;
                else if (point.y() >= (height()-corner))
                        pos = PositionBottomRight;
                else
                        pos = PositionRight;
        } else {
                // inside the frame
                pos = PositionCenter;
        }

        return pos;
}

//////////////////////////////////////////////////////////////////////////////
// eventFilter()
// -------------
// Event filter

bool GLOceanClient::eventFilter(QObject *obj, QEvent *e)
{

        if (obj != widget())
                return false;

        switch (e->type()) {
        case QEvent::MouseButtonDblClick: {
                        mouseDoubleClickEvent(static_cast<QMouseEvent *>(e));
                        return true;
                }
        case QEvent::MouseButtonRelease: {
                        mouseButtonReleaseEvent(static_cast<QMouseEvent *>(e));
                        return true;
                }
        case QEvent::MouseButtonPress: {
                        if( ::factory->glInitialized() ) {
                                for (int n=0; n<ButtonTypeCount; n++) {
                                        if(button[n]) {
                                                if(button[n]->geometry().contains( static_cast<QMouseEvent *>(e)->pos() )
                                                  ) {
                                                        button[n]->setDown(true);
                                                        return true;
                                                }
                                        }
                                }
                        }
                        processMousePressEvent(static_cast<QMouseEvent *>(e));
                        return true;
                }
        case QEvent::MouseMove: {
                        mouseMoveEvent(static_cast<QMouseEvent *>(e));
                        return true;
                }
        case QEvent::Paint: {
                        paintEvent(static_cast<QPaintEvent *>(e));
                        return true;
                }
        case QEvent::Resize: {
                        resizeEvent(static_cast<QResizeEvent *>(e));
                        return true;
                }
        case QEvent::Show: {
                        showEvent(static_cast<QShowEvent *>(e));
                        return true;
                }
        case QEvent::Hide: {
                        hideEvent(static_cast<QHideEvent *>(e));
                        return false;
                }
        case QEvent::Enter: {
                        enterEvent();
                        return true;
                }
        case QEvent::Leave: {
                        leaveEvent();
                        return false;
                }
        default: {
                        return false;
                }
        }

        return false;
}

//////////////////////////////////////////////////////////////////////////////
// mouseDoubleClickEvent()
// -----------------------
// Doubleclick on title

void GLOceanClient::mouseDoubleClickEvent(QMouseEvent *e)
{
        if (titlebar_->geometry().contains(e->pos()))
                titlebarDblClickOperation();
}

void GLOceanClient::mouseMoveEvent(QMouseEvent *e)
{
        bool flagged = false;

        if( ::factory->glInitialized() ) {
                for (int n=0; n<ButtonTypeCount; n++) {
                        if(button[n]) {
                                if(button[n]->geometry().contains( e->pos() ) ) {
                                        glGlobe->setPicked(n);
                                        button[n]->setHover(true);
                                        button[n]->repaint(false);
                                        flagged = true;
                                } else {
                                        if(!flagged)
                                                glGlobe->setPicked(-1);
                                        button[n]->setHover(false);
                                        button[n]->setDown(false);
                                        button[n]->repaint(false);
                                }
                        }
                }
        }
}

void GLOceanClient::mouseButtonReleaseEvent(QMouseEvent *e)
{
        if( ::factory->glInitialized() ) {
                for (int n=0; n<ButtonTypeCount; n++) {
                        if(button[n]) {
                                if(button[n]->geometry().contains( e->pos() ) ) {
                                        button[n]->setHover(false);
                                        button[n]->setDown(false);
                                        button[n]->animateClick();
                                        return;
                                }
                        }
                }
        }
}

//////////////////////////////////////////////////////////////////////////////
// paintEvent()
// ------------
// Repaint the window

void GLOceanClient::paintEvent(QPaintEvent*)
{
        if (!::factory->initialized())
                return;

        if(!::factory->fullFrame() ) {

                int l = ::factory->frameSize()-2;
                int b = (::factory->frameSize()-2)*2;

                // choose a color from scheme
                QColorGroup group = options()->colorGroup(KDecoration::ColorTitleBar, isActive());
                QPainter painter(widget());
                QRect frame(widget()->rect());

                // outline the frame
                frame.setRect(  frame.x(), ::factory->titleSize(),
                                frame.width(),
                                frame.height() - ::factory->titleSize());
                painter.setPen(group.dark());
                painter.drawRect(frame);
                // frame thickness titlebar color
                frame.setRect(frame.x() + 1, frame.y() + 1,
                              frame.width() - 2,
                              frame.height() - 2);
                painter.fillRect(frame,options()->color( KDecoration::ColorFrame, isActive()));
                // client frame default frame color
                frame.setRect(frame.x() + l, frame.y() + l,
                              frame.width() - b ,
                              frame.height() - b - ::factory->frameSize() );
                painter.setPen(group.dark());
                painter.drawRect(frame);
                // local temp right and bottom
                int t = height();
                int r = width();
                int cornersFlag = ::factory->cornerFlags();

                if(cornersFlag > 0) {
                        painter.setPen(group.dark());

                        // Draw edge of bottom-left corner inside the area removed by the mask.
                        if(cornersFlag & 4) {
                                painter.drawPoint(1, t - 5);
                                painter.drawPoint(1, t - 4);
                                painter.drawPoint(2, t - 3);
                                painter.drawPoint(3, t - 2);
                                painter.drawPoint(4, t - 2);
                        }
                        // Draw edge of bottom-right corner inside the area removed by the mask.
                        if(cornersFlag & 8) {
                                painter.drawPoint(r - 2, t - 5);
                                painter.drawPoint(r - 2, t - 4);
                                painter.drawPoint(r - 3, t - 3);
                                painter.drawPoint(r - 4, t - 2);
                                painter.drawPoint(r - 5, t - 2);
                        }
                }
        }
}

//////////////////////////////////////////////////////////////////////////////
// updateMask()
// ------------
// update the frame mask
void GLOceanClient::updateMask()
{
        int cornersFlag = ::factory->cornerFlags();
        int r(width());
        int b(height());
        QRegion mask;
        mask=QRegion(widget()->rect());

        if ( (cornersFlag == 0) || (!options()->moveResizeMaximizedWindows() && maximizeMode() & MaximizeFull ) ) {
                setMask(mask);
                return;
        }
        // Remove top-left corner.
        if(cornersFlag & 1) {
                mask -= QRegion(0, 0, 5, 1);
                mask -= QRegion(0, 1, 3, 1);
                mask -= QRegion(0, 2, 2, 1);
                mask -= QRegion(0, 3, 1, 2);
        }
        // Remove top-right corner.
        if(cornersFlag & 2) {
                mask -= QRegion(r - 5, 0, 5, 1);
                mask -= QRegion(r - 3, 1, 3, 1);
                mask -= QRegion(r - 2, 2, 2, 1);
                mask -= QRegion(r - 1, 3, 1, 2);
        }
        // Remove bottom-left corner.
        if(cornersFlag & 4) {
                mask -= QRegion(0, b - 5, 1, 3);
                mask -= QRegion(0, b - 3, 2, 1);
                mask -= QRegion(0, b - 2, 3, 1);
                mask -= QRegion(0, b - 1, 5, 1);
        }
        // Remove bottom-right corner.
        if(cornersFlag & 8) {
                mask -= QRegion(r - 5, b - 1, 5, 1);
                mask -= QRegion(r - 3, b - 2, 3, 1);
                mask -= QRegion(r - 2, b - 3, 2, 1);
                mask -= QRegion(r - 1, b - 5, 1, 2);
        }
        /// this should be alpha blend over client area
        /// 	if(::factory->fullFrame() ) {
        ///                 QRect frame(widget()->rect());
        /// 		// outline the frame
        /// 		frame.setRect(  frame.x(), ::factory->titleSize(),
        /// 			frame.width(),
        /// 			frame.height() - ::factory->titleSize());
        /// 		// frame thickness
        /// 		frame.setRect(frame.x() + 1, frame.y() + 1,
        /// 			frame.width() - 2,
        /// 			frame.height() - 2);
        /// 		// client frame
        /// 		frame.setRect(frame.x() + ::factory->frameSize()-2, frame.y() + ::factory->frameSize()-2,
        /// 			frame.width() - (::factory->frameSize()-2)*2 ,
        /// 			frame.height() - (::factory->frameSize()-2)*2 - ::factory->frameSize() );
        /// 		mask -= QRegion(frame.x(), frame.y(), frame.width(), frame.height());
        /// 	}
        setMask(mask);
}

//////////////////////////////////////////////////////////////////////////////
// resizeEvent()
// -------------
// Window is being resized

void GLOceanClient::resizeEvent(QResizeEvent *)
{
        if (widget()->isShown()) {
                QRegion region = widget()->rect();
                region = region.subtract(titlebar_->geometry());
                widget()->erase(region);
                if(::factory->glInitialized()) {
                        glGlobe->setPicked(-1);
                        glGlobe->resize( widget()->width(),
                                         (::factory->fullFrame() )  ? widget()->height() : ::factory->titleSize() );
                }
        }
        updateMask();
}

//////////////////////////////////////////////////////////////////////////////
// showEvent()
// -----------
// Window is being shown

void GLOceanClient::showEvent(QShowEvent *)
{
        if(::factory->glInitialized()) {
                glGlobe->setHasHelp(providesContextHelp());
                glGlobe->setHasMin(isMinimizable());
                glGlobe->setHasMax(isMaximizable());
                glGlobe->setHasClose(isCloseable());
                glGlobe->setHasShade(isShadeable());
                glGlobe->setPicked(-1);
                glGlobe->start(::factory->frameDel());
                started_ = true;
                if( isActive() && !started_ ) {
                        glGlobe->start(::factory->frameDel());
                        started_ = true;
                }
                if( !isActive() && started_ && ::factory->activeOnly() ) {
                        glGlobe->stop();
                        started_ = false;
                }
        }
        for (int n=0; n<ButtonTypeCount; n++) {
                if(button[n]) {
                        button[n]->setHover(false);
                        button[n]->setDown(false);
                }
        }
        widget()->repaint();
}

//////////////////////////////////////////////////////////////////////////////
// hideEvent()
// -----------
// Window is being hidden
void GLOceanClient::hideEvent(QHideEvent *)
{
        if(::factory->glInitialized()) {
                glGlobe->setPicked(-1);
                glGlobe->stop();
        }
}

//////////////////////////////////////////////////////////////////////////////
// maxButtonPressed()
// -----------------
// Max button was pressed

void GLOceanClient::maxButtonPressed()
{
        if (button[ButtonMax]) {
#if KDE_IS_VERSION(3, 3, 0)
                maximize(button[ButtonMax]->lastMousePress());
#else

                switch (button[ButtonMax]->lastMousePress()) {
                case MidButton:
                        maximize(maximizeMode() ^ MaximizeVertical);
                        break;
                case RightButton:
                        maximize(maximizeMode() ^ MaximizeHorizontal);
                        break;
                default:
                        (maximizeMode() == MaximizeFull) ? maximize(MaximizeRestore)
                        : maximize(MaximizeFull);
                }
#endif

        }
}

//////////////////////////////////////////////////////////////////////////////
// shadeButtonPressed()
// -----------------
// Shade button was pressed

void GLOceanClient::shadeButtonPressed()
{
        if (button[ButtonShade]) {
                setShade( !isSetShade());
        }
}

//////////////////////////////////////////////////////////////////////////////
// aboveButtonPressed()
// -----------------
// Above button was pressed

void GLOceanClient::aboveButtonPressed()
{
        if (button[ButtonAbove]) {
                setKeepAbove( !keepAbove());
        }
}

//////////////////////////////////////////////////////////////////////////////
// belowButtonPressed()
// -----------------
// Below button was pressed

void GLOceanClient::belowButtonPressed()
{
        if (button[ButtonBelow]) {
                setKeepBelow( !keepBelow());
        }
}

//////////////////////////////////////////////////////////////////////////////
// menuButtonPressed()
// -------------------
// Menu button was pressed (popup the menu)

void GLOceanClient::menuButtonPressed()
{
        if (button[ButtonMenu]) {
                QPoint p(button[ButtonMenu]->rect().bottomLeft().x(),
                         button[ButtonMenu]->rect().bottomLeft().y());
                KDecorationFactory* f = factory();
                showWindowMenu(button[ButtonMenu]->mapToGlobal(p));
                if (!f->exists(this))
                        return; // decoration was destroyed
                button[ButtonMenu]->setDown(false);
        }
}

#include "gloceanclient.moc"

