//////////////////////////////////////////////////////////////////////////////
// gloceanclient.h
// -------------------
// GLOcean window decoration for KDE 3.3+
// -------------------
// Copyright (c) 2004 R Paquette <bitwit@digilanti.org>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//////////////////////////////////////////////////////////////////////////////

#ifndef GLOCEANCLIENT_H
#define GLOCEANCLIENT_H
#include <GL/glx.h>
#include <qgl.h>

#include <qimage.h>
#include <qlayout.h>
#include <qbutton.h>
#include <kdecoration.h>
#include <kdecorationfactory.h>
#include <qmessagebox.h>
#include "globe.h"

class QSpacerItem;
class QPoint;
class GlobeWidget;
class GLFont;

namespace GLOcean
{

class GLOceanClient;

enum ButtonType {
        ButtonHelp=0,
        ButtonMax,
        ButtonMin,
        ButtonClose,
        ButtonMenu,
        ButtonSticky,
        ButtonAbove,
        ButtonBelow,
        ButtonShade,
        ButtonTypeCount
};

// GLOceanFactory /////////////////////////////////////////////////////////////

class GLOceanFactory: public KDecorationFactory
{
public:
        GLOceanFactory();
        virtual ~GLOceanFactory();
        virtual KDecoration *createDecoration(KDecorationBridge *b);
        virtual bool reset(unsigned long changed);
        void init();

        inline static bool initialized()
        {
                return initialized_;
        };
        inline static Qt::AlignmentFlags GLOceanFactory::titleAlign()
        {
                return titlealign_;
        };
        inline static int cornerFlags()
        {
                return cornerflags_;
        };
        inline static int titleSize()
        {
                return titlesize_;
        };
        inline static int buttonSize()
        {
                return buttonsize_;
        };
        inline static int frameSize()
        {
                return framesize_+4;
        };
        inline static int frameDel()
        {
                return framedelay_;
        };
        inline static bool titleColors()
        {
                return titlecolors_;
        };
        inline static bool hideButtons()
        {
                return hidebuttons_;
        };
        inline static int alphaBlend()
        {
                return alphablend_;
        };
        inline static int buttonBlend()
        {
                return buttonblend_;
        };
        inline static bool fadeText()
        {
                return fadetext_;
        };
        inline static bool activeOnly()
        {
                return activeonly_;
        };
        inline static bool animateButtons()
        {
                return animate_;
        };
        inline static bool glInitialized()
        {
                return glinitialized_;
        };
        inline static int symbolType()
        {
                return symboltype_;
        };
        inline static int animType()
        {
                return animtype_;
        };
        inline static bool fullFrame()
        {
                return fullframe_;
        };
        inline static bool hwRender()
        {
                return hwrender_;
        };
        inline static bool hasShadow()
        {
                return shadow_;
        };
        inline static GLXContext glxContext()
        {
                return glxcontext;
        };
        inline static QImage backFile()
        {
                return backfile;
        };
        inline static QImage buttFile()
        {
                return buttfile;
        };

        static GLXContext glxcontext;
protected:
        static QImage backfile;
        static QImage buttfile;
private:
        bool readConfig();
private:
        static bool glinitialized_;
        static bool initialized_;
        static Qt::AlignmentFlags titlealign_;
        static int cornerflags_;
        static int titlesize_;
        static int buttonsize_;
        static int framesize_;
        static int framedelay_;
        static bool titlecolors_;
        static bool hidebuttons_;
        static int alphablend_;
        static int buttonblend_;
        static bool fadetext_;
        static bool activeonly_;
        static bool animate_;
        static int symboltype_;
        static int animtype_;
        static bool fullframe_;
        static bool hwrender_;
        static bool shadow_;
};

// GLOceanButton //////////////////////////////////////////////////////////////

class GLOceanButton : public QButton
{
public:
        GLOceanButton(GLOceanClient *parent=0, const char *name=0,
                      const QString &tip=NULL,
                      ButtonType type=ButtonHelp,
                      const unsigned char *bitmap=0);
        ~GLOceanButton();

        void setBitmap(const unsigned char *bitmap);
        QSize sizeHint() const;
        inline Qt::ButtonState lastMousePress() const
        {
                return lastmouse_;
        };
        inline void reset()
        {
                repaint(false);
        };

        inline void setHover( bool f )
        {
                hover_ = f;
                emit QEvent::Enter;
        };

private:
        void mousePressEvent(QMouseEvent *e);
        void mouseReleaseEvent(QMouseEvent *e);
        void drawButton(QPainter *painter);

private:
        GLOceanClient *client_;
        ButtonType type_;
        QBitmap *deco_;
        ButtonState lastmouse_;
        bool hover_;
        static const char * const default_xpm[];
};

// GLOceanClient //////////////////////////////////////////////////////////////

class GLOceanClient : public KDecoration
{
        Q_OBJECT
public:
        GLOceanClient(KDecorationBridge *b, KDecorationFactory *f);
        virtual ~GLOceanClient();

        virtual void init();

        virtual void activeChange();
        virtual void desktopChange();
        virtual void captionChange();
        virtual void iconChange();
        virtual void maximizeChange();
        virtual void shadeChange();

        virtual void borders(int &l, int &r, int &t, int &b) const;
        virtual void resize(const QSize &size);
        virtual QSize minimumSize() const;
        virtual Position mousePosition(const QPoint &point) const;
        GlobeWidget *glGlobe;

private:
        void setButtonBits( const unsigned char bits[][8] );
        void addButtons(QBoxLayout* layout, const QString& buttons);
        bool eventFilter(QObject *obj, QEvent *e);
        void mouseDoubleClickEvent(QMouseEvent *e);
        void mouseMoveEvent(QMouseEvent *e);
        void mouseButtonReleaseEvent(QMouseEvent *e);
        void paintEvent(QPaintEvent *e);
        void resizeEvent(QResizeEvent *);
        void showEvent(QShowEvent *);
        void hideEvent(QHideEvent *);
        void updateMask();
        void enterEvent();
        void leaveEvent();

private slots:
        void maxButtonPressed();
        void menuButtonPressed();
        void aboveButtonPressed();
        void belowButtonPressed();
        void shadeButtonPressed();
        void keepAboveChange(bool);
        void keepBelowChange(bool);

private:
        GLOceanButton *button[ButtonTypeCount];
        QSpacerItem *titlebar_;
        bool started_;
        int handlebar;
        const unsigned char *close_bits      ;
        const unsigned char *help_bits       ;
        const unsigned char *max_bits        ;
        const unsigned char *min_bits        ;
        const unsigned char *restore_bits    ;
        const unsigned char *sticky_on_bits  ;
        const unsigned char *sticky_off_bits ;
        const unsigned char *above_on_bits   ;
        const unsigned char *above_off_bits  ;
        const unsigned char *below_on_bits   ;
        const unsigned char *below_off_bits  ;
        const unsigned char *shade_on_bits   ;
        const unsigned char *shade_off_bits  ;

        static const unsigned char default_bits[][8];
        static const unsigned char plastic_bits[][8];
        static const unsigned char kstep_bits[][8];
        static const unsigned char control_bits[][8];
};

//// close_bits[]
//// help_bits[]
//// max_bits[]
//// min_bits[]
//// restore_bits[]
//// sticky_off_bits[]
//// sticky_on_bits[]
//// above_off_bits[]
//// above_on_bits[]
//// below_off_bits[]
//// below_on_bits[]
//// shade_off_bits[]
//// shade_on_bits[]

const unsigned char GLOceanClient::default_bits[][8] =
        {
                {
                        0x00, 0x66, 0x7e, 0x3c, 0x3c, 0x7e, 0x66, 0x00
                },
                { 0x7e, 0x7e, 0x60, 0x78, 0x78, 0x00, 0x18, 0x18 },
                { 0x00, 0x18, 0x3c, 0x7e, 0xff, 0xff, 0x00, 0x00 },
                { 0x00, 0x00, 0xff, 0xff, 0x7e, 0x3c, 0x18, 0x00 },
                { 0x00, 0x02, 0x06, 0x0e, 0x1e, 0x3e, 0x7e, 0x00 },
                { 0x00, 0x18, 0x18, 0x7e, 0x7e, 0x18, 0x18, 0x00 },
                { 0x00, 0x00, 0x00, 0x7e, 0x7e, 0x00, 0x00, 0x00 },
                { 0x7e, 0x18, 0x7e, 0x3c, 0x18, 0x00, 0x00, 0x00 },
                { 0x18, 0x3c, 0x7e, 0x18, 0x7e, 0x00, 0x00, 0x00 },
                { 0x00, 0x00, 0x00, 0x18, 0x3c, 0x7e, 0x18, 0x7e },
                { 0x00, 0x00, 0x00, 0x7e, 0x18, 0x7e, 0x3c, 0x18 },
                { 0xff, 0xff, 0x81, 0xa5, 0x81, 0xa5, 0x81, 0xff },
                { 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
        };

const unsigned char GLOceanClient::plastic_bits[][8] =
        {
                {
                        0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81
                },
                { 0x1c, 0x36, 0x22, 0x30, 0x18, 0x08, 0x00, 0x08 },
                { 0xff, 0xff, 0x81, 0x81, 0x81, 0x81, 0x81, 0xff },
                { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff },
                { 0xfc, 0xfc, 0x84, 0x9f, 0x9f, 0xf1, 0x11, 0x1f },
                { 0x3c, 0x42, 0x81, 0x81, 0x81, 0x81, 0x42, 0x3c },
                { 0x00, 0x3c, 0x66, 0x5a, 0x5a, 0x66, 0x3c, 0x00 },
                { 0x00, 0x00, 0x18, 0x24, 0x42, 0xff, 0x00, 0x00 },
                { 0x00, 0x00, 0x18, 0x3c, 0x7e, 0xff, 0x00, 0x00 },
                { 0x00, 0x00, 0xff, 0x42, 0x24, 0x18, 0x00, 0x00 },
                { 0x00, 0x00, 0xff, 0x7e, 0x3c, 0x18, 0x00, 0x00 },
                { 0xff, 0xff, 0x81, 0xff, 0x00, 0x18, 0x3c, 0x7e },
                { 0xff, 0xff, 0x81, 0x99, 0xbd, 0xff, 0x81, 0xff }
        };

const unsigned char GLOceanClient::kstep_bits[][8] =
        {
                {
                        0xc3, 0xe7, 0x7e, 0x3c, 0x3c, 0x7e, 0xe7, 0xc3
                },
                { 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18 },
                { 0x18, 0x3c, 0x7e, 0x00, 0xff, 0x42, 0x24, 0x18 },
                { 0xff, 0xff, 0x81, 0x81, 0x81, 0x81, 0x81, 0xff },
                { 0xfc, 0xf8, 0xf1, 0xe3, 0xc5, 0x89, 0x11, 0x3f },
                { 0x18, 0x18, 0x18, 0xff, 0xff, 0x18, 0x18, 0x18 },
                { 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00 },
                { 0x18, 0x3c, 0x7e, 0x00, 0xff, 0xff, 0x00, 0x00 },
                { 0xff, 0xff, 0x00, 0x7e, 0x3c, 0x18, 0x00, 0x00 },
                { 0x00, 0x00, 0xff, 0xff, 0x00, 0x7e, 0x3c, 0x18 },
                { 0x00, 0x00, 0x18, 0x3c, 0x7e, 0x00, 0xff, 0xff },
                { 0xff, 0xff, 0xc3, 0xff, 0xff, 0x00, 0x00, 0x00 },
                { 0xff, 0xff, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xff }
        };

const unsigned char GLOceanClient::control_bits[][8] =
        {
                {
                        0x81, 0xc3, 0x66, 0x3c, 0x3c, 0x66, 0xc3, 0x81
                },
                { 0x3c, 0x46, 0x46, 0x40, 0x20, 0x10, 0x18, 0x18 },
                { 0x00, 0x7e, 0x42, 0x42, 0x42, 0x42, 0x7e, 0x00 },
                { 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00 },
                { 0xff, 0x89, 0x89, 0x89, 0xf9, 0x81, 0x81, 0xff },
                { 0x00, 0x3c, 0x66, 0x5a, 0x5a, 0x66, 0x3c, 0x00 },
                { 0x10, 0x30, 0x70, 0x5e, 0x7f, 0x70, 0x30, 0x10 },
                { 0xff, 0x99, 0xbd, 0xff, 0x99, 0x99, 0x81, 0xff },
                { 0xff, 0x00, 0x18, 0x3c, 0x7e, 0x18, 0x18, 0x18 },
                { 0xff, 0x81, 0x99, 0x99, 0xff, 0xbd, 0x99, 0xff },
                { 0x18, 0x18, 0x18, 0x7e, 0x3c, 0x18, 0x00, 0xff },
                { 0xff, 0x81, 0xff, 0x18, 0x3c, 0x7e, 0x18, 0xa5 },
                { 0xff, 0x81, 0xff, 0xbd, 0x99, 0x81, 0x81, 0xff }
        };

} // namespace GLOcean

#endif // GLOCEANCLIENT_H
